/*
 * $Id: DocLint.java,v 1.1 2000/07/05 18:14:39 ernst Exp $
 *
 * Copyright (C) 2000  Ernst de Haan
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
package com.jollem.doclint;

import com.sun.javadoc.*;
import java.util.*;

/**
 * Doclet that checks that there is information available for every package,
 * class, interface and member.
 *
 * @version $Revision: 1.1 $
 * @author Ernst de Haan
 *         (<A href="mailto:ernst@jollem.com">ernst@jollem.com</A>)
 */
public class DocLint extends Doclet {

   protected static RootDoc    _rootDoc;

   protected static PackageDoc _packageDoc;
   protected static String     _packageName;

   protected static ClassDoc _classDoc;
   protected static String   _className;

   protected static ExecutableMemberDoc _executableMemberDoc;
   protected static String              _executableMemberName;
   protected static String              _executableMemberSignature;

   protected static FieldDoc _fieldDoc;
   protected static String   _fieldName;

   /**
    * Run this doclet for the specified packages or classes.
    *
    * @param rootDoc
    *    The information for this run of <EM>javadoc</EM>.
    *
    * @return
    *    On success, <CODE>true</CODE>, on failure: <CODE>false</CODE>.
    */
   public static boolean start(RootDoc rootDoc) {

      // Set current RootDoc
      _rootDoc = rootDoc;

      // Perform checks
      checkPackages(rootDoc.specifiedPackages());
      checkClasses(rootDoc.specifiedClasses());

      // Unset current RootDoc
      _rootDoc = null;

      return true;
   }

   protected static void checkPackages(PackageDoc[] packages) {
      int packageCount = (packages==null) ? 0 : packages.length;
      for (int i=0; i<packageCount; i++) {

         // Set current PackageDoc
         _packageDoc = packages[i];
         _packageName = _packageDoc.name();

         // Get details
         String     comment     = _packageDoc.commentText();
         ClassDoc[] classes     = _packageDoc.allClasses();

         // Perform checks
         if (isEmptyString(comment)) {
            missingPackageComment();
         }
         checkClasses(classes);
      }

      // Unset current PackageDoc
      _packageDoc = null;
      _packageName = null;
   }

   protected static void checkClasses(ClassDoc[] classes) {
      int classCount = (classes==null) ? 0 : classes.length;
      String packageName = (_packageDoc==null) ? "<none>" : _packageDoc.name();
      for (int i=0; i<classCount; i++) {

         // Set current ClassDoc
         _classDoc = classes[i];
         _className = _classDoc.name();

         // Get details
         String comment                = _classDoc.commentText();
         ConstructorDoc[] constructors = _classDoc.constructors();
         MethodDoc[] methods           = _classDoc.methods();
         FieldDoc[] fields             = _classDoc.fields();

         // Perform checks
         if (isEmptyString(comment)) {
            missingClassComment();
         }
         checkExecutableMembers(constructors);
         checkExecutableMembers(methods);
         checkFields(fields);
      }

      // Unset current ClassDoc
      _classDoc = null;
      _className = null;
   }

   protected static void checkFields(FieldDoc[] fields) {
      int fieldCount = (fields==null) ? 0 : fields.length;
      for (int i=0; i<fieldCount; i++) {

         // Set current field
         _fieldDoc = fields[i];
         _fieldName = _fieldDoc.name();

         // Get details
         String comment = _fieldDoc.commentText();

         // Perform checks
         if (isEmptyString(comment)) {
            _rootDoc.printWarning("No comments for " + _packageName +
                                  '.' + _className +
                                  ", field " + _fieldName);
         }
         
         checkField();

         // Unset current field
         _fieldDoc = null;
         _fieldName = null;
      }
   }

   protected final static void checkField() {
      if (_fieldDoc.isStatic()) {
         checkStaticField();
      } else {
         checkInstanceField();
      }
   }

   protected static void checkStaticField() {
   }

   protected static void checkInstanceField() {
   }

   protected static void checkExecutableMembers(
      ExecutableMemberDoc[] members) {

      int memberCount = (members==null) ? 0 : members.length;
      for (int i=0; i<memberCount; i++) {

         // Set current ExecutableMemberDoc
         _executableMemberDoc = members[i];
         _executableMemberName =
            (_executableMemberDoc instanceof ConstructorDoc)
               ? "<init>"
               : _executableMemberDoc.name();
         _executableMemberSignature = _executableMemberDoc.flatSignature();

         // Get details
         String comment = _executableMemberDoc.commentText();

         // Perform checks
         if (isEmptyString(comment)) {
            missingExecutableMemberComment();
         }

         checkParameters();
         checkReturn();
         checkExceptions();
      }

      // Unset current MemberDoc
      _executableMemberDoc       = null;
      _executableMemberName      = null;
      _executableMemberSignature = null;
   }

   protected static void checkParameters() {
      Parameter[] params = _executableMemberDoc.parameters();
      ParamTag[]  tags   = _executableMemberDoc.paramTags();

      int paramCount = params.length;
      int tagCount   = tags.length;

      // Put all the tags in a map
      Map tagMap   = new HashMap(tagCount);
      for (int i=0; i<tagCount; i++) {
         ParamTag tag     = tags[i];
         String   name    = tag.parameterName();
         String   comment = tag.parameterComment();
         if (tagMap.containsKey(name)) {
            _rootDoc.printWarning("Multiple comments for " + _packageName +
                                  '.' + _className +
                                  '.' + _executableMemberName +
                                  '.' + _executableMemberSignature +
                                  ", parameter " + name);
         } else if (! isEmptyString(comment)) {
            tagMap.put(name, tag);
         }
      }

      // Check that every parameter is described by a tag
      for (int i=0; i<paramCount; i++) {
         Parameter param = params[i];
         String    name  = param.name();
         if (! tagMap.containsKey(name)) {
            _rootDoc.printWarning("No comments for " + _packageName +
                                  '.' + _className +
                                  '.' + _executableMemberName +
                                  '.' + _executableMemberSignature +
                                  ", parameter " + name);
         }
      }
      tagMap = null;

      // Put all the parameters in a map
      Map paramMap = new HashMap(paramCount);
      for (int i=0; i<paramCount; i++) {
         Parameter param = params[i];
         String    name  = param.name();
         paramMap.put(name, param);
      }

      // Check that every tag describes a parameter
      for (int i=0; i<tagCount; i++) {
         ParamTag tag = tags[i];
         String paramName = tag.parameterName();
         if (! paramMap.containsKey(paramName)) {
            _rootDoc.printWarning("Found comment for non-existent " +
                                  _packageName +
                                  '.' + _className +
                                  '.' + _executableMemberName +
                                  '.' + _executableMemberSignature +
                                  ", parameter " + paramName);
         }
      }
   }

   protected static void checkReturn() {
      if (! (_executableMemberDoc instanceof MethodDoc)) {
         return;
      }

      MethodDoc methodDoc = (MethodDoc) _executableMemberDoc;
      Type  returnType    = methodDoc.returnType();
      if ("void".equals(returnType.qualifiedTypeName())) {
         return;
      }

      Tag[] returnTags = methodDoc.tags("return");
      int tagCount = (returnTags==null) ? 0 : returnTags.length;
      if (tagCount == 0) {
         _rootDoc.printWarning("No comments for " + _packageName +
                               '.' + _className +
                               '.' + _executableMemberName +
                               '.' + _executableMemberSignature +
                               ", return value.");
      } else if (tagCount > 1) {
         _rootDoc.printWarning("Multiple comments for " + _packageName +
                               '.' + _className +
                               '.' + _executableMemberName +
                               '.' + _executableMemberSignature +
                               ", return value (" + tagCount + ").");
      }
   }

   protected static void checkExceptions() {
      ClassDoc[]  exceptions = _executableMemberDoc.thrownExceptions();
      ThrowsTag[] tags       = _executableMemberDoc.throwsTags();

      int exceptionCount = exceptions.length;
      int tagCount       = tags.length;

      // Put all the tags in a map
      Map tagMap   = new HashMap(tagCount);
      for (int i=0; i<tagCount; i++) {
         ThrowsTag tag  = tags[i];
         String    name = tag.exceptionName();
         tagMap.put(name, tag);
      }

      // Check that every exception is described by a tag
      for (int i=0; i<exceptionCount; i++) {
         ClassDoc exception = exceptions[i];
         String name = exception.name();
         if (! tagMap.containsKey(name)) {
            _rootDoc.printWarning("No comments for " + _packageName +
                                  '.' + _className +
                                  '.' + _executableMemberName +
                                  '.' + _executableMemberSignature +
                                  ", exception " + name);
         }
      }
      tagMap = null;

      // Put all the exceptions in a map
      Map exceptionMap = new HashMap(exceptionCount);
      for (int i=0; i<exceptionCount; i++) {
         ClassDoc exception = exceptions[i];
         String   name      = exception.name();
         exceptionMap.put(name, exception);
      }

      // Check that every tag describes an exception
      for (int i=0; i<tagCount; i++) {
         ThrowsTag tag = tags[i];
         String exceptionName = tag.exceptionName();
         if (! exceptionMap.containsKey(exceptionName)) {
            _rootDoc.printWarning("Found comment for non-existent " +
                                  _packageName +
                                  '.' + _className +
                                  '.' + _executableMemberName +
                                  '.' + _executableMemberSignature +
                                  ", exception " + exceptionName);
         }
      }
   }


   //-------------------------------------------------------------------------
   // handle different special conditions
   //-------------------------------------------------------------------------

   protected static void missingPackageComment() {
      _rootDoc.printWarning("No comments for " + _packageName);
   }

   protected static void missingClassComment() {
      _rootDoc.printWarning("No comments for " + _packageName +
                            '.' + _className);
   }

   protected static void missingExecutableMemberComment() {
      _rootDoc.printWarning("No comments for " + _packageName +
                            '.' + _className +
                            '.' + _executableMemberName +
                            _executableMemberSignature);
   }


   //-------------------------------------------------------------------------
   // utility functions
   //-------------------------------------------------------------------------

   protected final static boolean isEmptyString(String s) {
      if (s == null || s.trim().length() == 0) {
         return true;
      } else {
         return false;
      }
   }
}
